from datetime import datetime
import json
import os
import shutil
import tarfile
from urllib.parse import urljoin

import requests

from mswalletrpc import MultisigWallet


class BaseRPC():
    form = {}

    def get(self, request, *args, **kargs):
        raise NotImplementedError

    def post(self, request, *args, **kargs):
        raise NotImplementedError

    @classmethod
    def get_form_data(cls, data):
        res = {}
        for name, _type in cls.form.items():
            if name in data:
                value = data[name]
                if isinstance(value, _type):
                    res[name] = value
                else:
                    if value and isinstance(value, (tuple, list)):
                        value = value[0]
                    if _type == bool:
                        res[name] = value.lower() == 'true'
                    else:
                        try:
                            res[name] = _type(value)
                        except Exception as e:
                            print(f'ERROR: {e}')
        return res

    def success(self, result=True):
        return {'result': result}

    def error(self, errors='Server error'):
        return {'result': False, 'errors': errors if isinstance(errors, (list, tuple)) else [errors]}


class SettingsRPC(BaseRPC):
    form = {
        'daemon-address': str,
        'daemon-login': str,
        'stagenet': bool,
        'testnet': bool,
        'proxy': str,
        'timeout': int,
    }

    def get(self, request):
        return request.settings

    def post(self, request, data):
        changed = False
        for name in request.settings.keys():
            value = data.get(name)
            if value:
                if value != request.settings.get(name):
                    request.settings[name] = value
                    changed = True
            elif request.settings.get(name):
                request.settings[name] = None
                changed = True
        if changed:
            request.save_settings()
        return self.success()


class SettingsTestRPC(BaseRPC):
    def post(self, request, data=None):
        try:
            if request.settings['daemon-login']:
                username, passwd = request.settings['daemon-login'].split(':')
                auth = requests.auth.HTTPDigestAuth(username, passwd)
            else:
                auth = None
            proxy = request.settings.get('proxy')
            if proxy:
                proxy = f'socks5h://{request.settings["proxy"]}'
            resp = requests.post(
                urljoin(request.settings['daemon-address'], 'json_rpc'),
                headers={'Content-Type': 'application/json'},
                data=json.dumps({'jsonrpc': '2.0', 'id': 0, 'method': 'get_version'}),
                auth=auth,
                verify=False,
                proxies={
                    'http': proxy,
                    'https': proxy,
                },
                timeout=30
            )
            if resp.status_code == 200:
                return self.success()
            return self.error(resp.content.decode())
        except Exception as e:
            return self.error(str(e))


class WalletListRPC(BaseRPC):
    def get(self, request):
        wallets = request.get_wallets() or {}
        return self.success(list(wallets.values()))

    def post(self, request, data=None):
        request.cd_default_dir()
        wallet = MultisigWallet.generate_wallet(request.settings)
        wallet.turn_ms_on()
        wallet.prepare_multisig()
        return self.success(wallet.name)


class WalletBaseRPC(BaseRPC):
    def get_wallet(self, request, wallet_id):
        request.cd_default_dir()
        return request.get_wallet(wallet_id)


class WalletPrepareRPC(WalletBaseRPC):
    def get(self, request, wallet_id):
        wallet = self.get_wallet(request, wallet_id)
        wallet.rpc_stop()
        wallet.turn_ms_on()
        wallet.prepare_multisig()
        return self.success(wallet.name)


class WalletDetailRPC(WalletBaseRPC):
    def get(self, request, wallet_id):
        wallet = self.get_wallet(request, wallet_id)
        if wallet.finalized:
            state = 'finalized'
        elif wallet.key3:
            state = 'round3'
        elif wallet.key2:
            state = 'round2'
        elif wallet.key1:
            state = 'round1'
        else:
            state = 'new'
        return {
            'name': wallet.name,
            'label': wallet.label,
            'address': wallet.get_address(),
            'process': bool(wallet.process),
            'key1': wallet.key1,
            'key2': wallet.key2,
            'key3': wallet.key3,
            'balance': getattr(wallet, '_balance', None),
            'state': state,
            'is_blocked': wallet.is_blocked,
        }


class WalletLabelRPC(WalletBaseRPC):
    form = {'label': str}

    def post(self, request, wallet_id, data):
        wallet = self.get_wallet(request, wallet_id)
        if wallet.errors:
            return self.error(wallet.errors)
        wallet.set_label(data.get('label'))
        return self.success()


class WalletBalanceRPC(WalletBaseRPC):
    def get(self, request, wallet_id):
        wallet = self.get_wallet(request, wallet_id)
        wallet._balance = wallet.balance()
        if wallet.errors:
            return self.error(wallet.errors)
        return self.success(wallet._balance)


class WalletRescanRPC(WalletBaseRPC):
    def get(self, request, wallet_id):
        wallet = self.get_wallet(request, wallet_id)
        if wallet.process:
            if wallet.test_command():
                wallet.is_blocked = False
                return self.success('done')
            else:
                return self.error(['Still updating'])
        wallet.rescan()
        return self.success('started')


class WalletTransfersRPC(WalletBaseRPC):
    form = {
        'address': str,
        'amount': float,
    }

    def get(self, request, wallet_id):
        wallet = self.get_wallet(request, wallet_id)
        data = wallet.show_transfers()
        if wallet.errors:
            return self.error(wallet.errors)
        result = []
        if data:
            for item in data.get('in', []):
                result.append({
                    'txid': item['txid'],
                    'income': True,
                    'amount': item['amount'] / 10**12,
                    'timestamp': datetime.fromtimestamp(item['timestamp']).strftime('%Y-%m-%d %H:%M:%S'),
                    'confirmations': item['confirmations'],
                })
            for item in data.get('out', []):
                result.append({
                    'txid': item['txid'],
                    'income': False,
                    'amount': item['amount'] / 10**12,
                    'timestamp': datetime.fromtimestamp(item['timestamp']).strftime('%Y-%m-%d %H:%M:%S'),
                    'confirmations': item['confirmations'],
                })
        return self.success(sorted(result, key=lambda a: a['timestamp']))

    def post(self, request, wallet_id, data):
        wallet = self.get_wallet(request, wallet_id)
        addr = data.get('address')
        amount = data.get('amount')
        if addr and amount:
            data = wallet.transfer(addr, amount)
            if wallet.errors:
                return self.error(wallet.errors)
            return self.success(data)
        return self.error('Address or amount are not correct')


class WalletCreateRPC(WalletBaseRPC):
    form = {'keys': str}

    def action(self, wallet, keys):
        raise NotImplementedError

    def post(self, request, wallet_id, data):
        wallet = self.get_wallet(request, wallet_id)
        errors = []
        keys = []
        for k in (data.get('keys') or '').split():
            if len(k) != self.key_length or not k.startswith(self.key_startswith):
                errors.append(f'{k[:15]}..{k[-15:]}: This is not valid round{self.round_num} multisig key')
            if k not in keys and getattr(wallet, f'key{self.round_num}', None) != k:
                keys.append(k)
        if len(keys) == 2:
            self.action(wallet, keys)
            if wallet.errors:
                return self.error(wallet.errors)
            return self.success(self.result(wallet))
        if len(keys) > 2:
            errors.append('Too many keys, try again...')
        elif len(keys) < 2:
            errors.append(f'Need {2-len(keys)} more keys...')
        return self.error(errors)


class WalletMakeRPC(WalletCreateRPC):
    key_length = 189
    key_startswith = 'MultisigxV2R1'
    round_num = 1

    def action(self, wallet, keys):
        wallet.make_multisig(*keys)

    def result(self, wallet):
        return wallet.key2


class WalletExchangeRPC(WalletCreateRPC):
    key_length = 236
    key_startswith = 'MultisigxV2Rn'
    round_num = 2

    def action(self, wallet, keys):
        wallet.exchange_multisig_keys(*keys)

    def result(self, wallet):
        return wallet.key3


class WalletFinalizeRPC(WalletExchangeRPC):
    round_num = 3

    def action(self, wallet, keys):
        wallet.finalize_wallet(*keys)

    def result(self, wallet):
        return wallet.address


class WalletExportInfoRPC(WalletBaseRPC):
    def get(self, request, wallet_id):
        wallet = self.get_wallet(request, wallet_id)
        return self.success(wallet.export_multisig_info())


class WalletImportInfoRPC(WalletBaseRPC):
    form = {'info': str}

    def post(self, request, wallet_id, data):
        wallet = self.get_wallet(request, wallet_id)
        resp = wallet.import_multisig_info(data.get('info'))
        if resp:
            return self.success()
        return self.error(wallet.errors)


class WalletTxShowRPC(WalletBaseRPC):
    form = {'tx': str}

    def post(self, request, wallet_id, data):
        wallet = self.get_wallet(request, wallet_id)
        resp = wallet.show_multisig(data.get('tx'))
        if wallet.errors:
            return self.error(wallet.errors)
        if resp:
            result = {'recipients': []}
            for item in resp['desc'][0]['recipients']:
                result['recipients'].append({
                    'address': item['address'],
                    'amount': item['amount'] / 10**12,
                })
            result['fee'] = resp['desc'][0]['fee'] / 10**12
            return self.success(result)
        return self.error('Tx is not valid')


class WalletTxSignRPC(WalletBaseRPC):
    form = {'tx': str}

    def post(self, request, wallet_id, data):
        wallet = self.get_wallet(request, wallet_id)
        resp = wallet.sign_multisig(data.get('tx'))
        if wallet.errors:
            return self.error(wallet.errors)
        if resp:
            return self.success(resp)
        return self.error('Tx is not valid')


class WalletTxSubmitRPC(WalletBaseRPC):
    form = {'tx': str}

    def post(self, request, wallet_id, data):
        wallet = self.get_wallet(request, wallet_id)
        resp = wallet.submit_multisig(data.get('tx'))
        if wallet.errors:
            return self.error(wallet.errors)
        if resp:
            return self.success(resp)
        return self.error('Tx is not valid')


class WalletArchiveRPC(WalletBaseRPC):
    def post(self, request, wallet_id, data):
        wallet = self.get_wallet(request, wallet_id)
        request.pool.kill(wallet.name)
        request.cd_default_dir()
        if not os.path.isdir('archives'):
            os.mkdir('archives')
        fn = wallet.label or wallet.key1[:32]
        tar = tarfile.open(os.path.join('archives', f'{fn}.tgz'), 'w:gz')
        tar.add(os.path.join('wallets', wallet.name), arcname=fn)
        tar.close()
        shutil.rmtree(os.path.join('wallets', wallet.name))
        return self.success()


class WalletStopRPC(WalletBaseRPC):
    def post(self, request, wallet_id, data):
        wallet = self.get_wallet(request, wallet_id)
        request.pool.kill(wallet.name)
        return self.success()


class WalletRunClientRPC(WalletBaseRPC):
    def get(self, request, wallet_id):
        wallet = self.get_wallet(request, wallet_id)
        cmd = ' \\\n\t'.join(wallet.get_cli_command())
        path = wallet.wallet_path
        return self.success(f'cd {path}\n{cmd}')
