#!/usr/bin/python3

import json
import os
import platform
import shutil
import tarfile
import time
from urllib.parse import urljoin

import requests

from mswalletrpc import MultisigWallet


class Console():
    settings = {
        'daemon-address': 'localhost:18081',
        'daemon-login': '',
        'stagenet': False,
        'testnet': False,
        # 'proxy': '',  # disabled because now proxy is not working in monero-wallet-rpc
    }

    def __init__(self):
        self.default_path = os.path.abspath(os.curdir)
        try:
            self.settings = json.loads(open('settings.json').read())
        except Exception:
            ...
        if platform.system() == 'Windows':
            self.edit_app = 'notepad'
            self.clear_cmd = 'cls'
        else:
            self.edit_app = 'open'
            self.clear_cmd = 'clear'

    def save_settings(self):
        self.cd_default_dir()
        open('settings.json', 'w').write(json.dumps(self.settings))

    def clear(self):
        os.system(self.clear_cmd)

    def cd_default_dir(self):
        os.chdir(self.default_path)

    def color(self, text, color=None):
        if color == 'red':
            return f'\033[91m{text}\033[0m'
        if color == 'green':
            return f'\033[92m{text}\033[0m'
        if color == 'yellow':
            return f'\033[33m{text}\033[0m'
        return text

    def make_multisig(self, wallet):
        if wallet.key2:
            print('Multisig wallet is already created, now you need to exchange keys')
            return
        print('Paste round1 keys of other participants')
        keys = []
        while True:
            inp = input()
            if not inp:
                break
            for k in inp.split():
                if len(k) != 189 or not k.startswith('MultisigxV2R1'):
                    print('This is not valid round1 multisig key')
                elif k in keys:
                    print('The same key, send another one')
                elif wallet.key1 == k:
                    print('This is your key, send another one')
                else:
                    keys.append(k)
            if len(keys) > 2:
                print('Too many keys, try again...')
                self.make_multisig(wallet)
                break
            if len(keys) < 2:
                print(f'Need {2-len(keys)} more keys...')
                continue
            if len(keys) == 2:
                self.clear()
                print('Making multisig...')
                wallet.make_multisig(*keys)
                break

    def exchange_keys(self, wallet, round_num):
        if wallet.finalized:
            print('Multisig wallet is already finalized and ready to receive XMR')
            return
        print(f'Paste round{round_num} keys of other participants')
        ownkey = wallet.key3 if round_num == 3 else wallet.key2
        keys = []
        while True:
            inp = input()
            if not inp:
                break
            for k in inp.split():
                if len(k) != 236 or not k.startswith('MultisigxV2Rn'):
                    print(f'This is not valid round{round_num} multisig key')
                elif k in keys:
                    print('The same key, send another one')
                elif ownkey == k:
                    print('This is your key, send another one')
                else:
                    keys.append(k)
            if len(keys) > 2:
                print('Too many keys, try again...')
                self.exchange_keys(wallet, round_num)
                break
            if len(keys) < 2:
                print(f'Need {2-len(keys)} more keys...')
                continue
            if len(keys) == 2:
                self.clear()
                if round_num == 3:
                    print('Finalize multisig...')
                    wallet.finalize_wallet(*keys)
                else:
                    print('Exchanging keys...')
                    wallet.exchange_multisig_keys(*keys)
                break

    def import_multisig(self, wallet):
        print('Paste hex of multisig info')
        while True:
            inp = input()
            if not inp:
                break
            self.clear()
            print('Importing multisig...')
            data = wallet.import_multisig_info(inp)
            self.clear()
            if data:
                print(f'Imported successfully: {data}')
            else:
                print('Could not receive outputs')
            break

    def transfer(self, wallet):
        print('Paste XMR address and amount separated by space to make transaction')
        while True:
            inp = input()
            if not inp:
                break
            data = inp.split(' ')
            if len(data) != 2:
                print('Wrong input, try again...')
                continue
            if len(data[0]) != 95:
                print('This is not valid XMR address, try again...')
                continue
            addr = data[0]
            try:
                amount = float(data[1])
            except Exception:
                print('This is not valid amount, try again...')
                continue
            data = wallet.transfer(addr, amount)
            if data:
                self.clear()
                print(f'Hex of transaction:{os.linesep}{data}')
            break

    def save_multisig_tx(self, wallet):
        fn = os.path.abspath(os.path.join(wallet.wallet_path, 'ms_tx_hex.txt'))
        if not os.path.exists(fn):
            open(fn, 'w+')
        os.system(f'{self.edit_app} {fn}')

    def multisig_tx(self, wallet, method):
        fn = os.path.abspath(os.path.join(wallet.wallet_path, 'ms_tx_hex.txt'))
        while True:
            if os.path.exists(fn):
                data = open(fn).read().strip()
            else:
                print('Paste absolute filename of multisig tx hex data')
                inp = input()
                if not inp:
                    break
                if not os.path.exists(inp):
                    print('File not found, try again...')
                    continue
                data = open(inp).read().strip()
            return getattr(wallet, method)(data)

    def show_multisig(self, wallet):
        self.multisig_tx(wallet, 'show_multisig')

    def sign_multisig(self, wallet):
        data = self.multisig_tx(wallet, 'sign_multisig')
        self.clear()
        if data:
            fn = os.path.abspath(os.path.join(wallet.wallet_path, 'ms_tx_signed_hex.txt'))
            open(fn, 'w').write(data)
            print('Transaction signed, now you can submit it')

    def submit_multisig(self, wallet):
        fn = os.path.abspath(os.path.join(wallet.wallet_path, 'ms_tx_signed_hex.txt'))
        if not os.path.exists(fn):
            print('Signed transaction not found')
            return
        data = wallet.submit_multisig(open(fn).read().strip())
        self.clear()
        if data:
            print(f'Transaction {data} submited')

    def set_label(self, wallet):
        print('Write down label for wallet or empty line to cancel')
        label = input()
        if label:
            wallet.set_label(label)
            self.clear()

    def wallet_detail(self, wallet):
        while True:
            info = [
                f'Wallet{wallet.name}: {wallet.label or "No label"}',
                self.color(f'Address: {wallet.get_address()}', 'green' if wallet.finalized else None),
            ]
            if wallet.process:
                info.append(self.color('RPC: running', 'green'))
            else:
                info.append('RPC: stopped')
            next_action = 1
            if wallet.key1:
                info.append(f'Round1 key: {wallet.key1}')
                if wallet.key2:
                    info.append(f'Round2 key: {wallet.key2}')
                    if wallet.key3:
                        info.append(f'Round3 key: {wallet.key3}')
                        if wallet.finalized:
                            info.append(self.color('Finalized and ready to receive XMR...', 'green'))
                            next_action = 4
                        else:
                            next_action = 3
                            info.append(self.color('Need to finalize wallet...', 'red'))
                    else:
                        next_action = 2
                        info.append(self.color('Need to exchange keys...', 'yellow'))
                else:
                    next_action = 1
                    info.append(self.color('Need to make multisig...', 'yellow'))
            info = os.linesep.join(info)
            print(f'''
{info}

Choose action:
{self.color('1. Make multisig wallet', next_action == 1 and 'green')}
{self.color('2. Exchange keys', next_action == 2 and 'green')}
{self.color('3. Finalize wallet', next_action == 3 and 'green')}
4. Export multisig info
5. Import multisig info

6. Transfer
7. Save transaction to file
8. Show transaction
9. Sign transaction
10. Submit transaction

11. Show balance
12. Show transfers
13. Rescan blockchain
14. Set label
15. Open console

20. Archive and delete

0. Back
            ''')
            action = input()
            self.clear()
            if action == '0':
                self.cd_default_dir()
                if wallet.process:
                    self.clear()
                    wallet.rpc_stop()
                    del wallet
                break
            if action == '1':
                self.make_multisig(wallet)
            elif action == '2':
                self.exchange_keys(wallet, 2)
            elif action == '3':
                self.exchange_keys(wallet, 3)
            elif action == '4':
                self.clear()
                print('Exporting multisig info...')
                data = wallet.export_multisig_info()
                self.clear()
                print(f'Hex of exmported multisig info:{os.linesep}{data}')
            elif action == '5':
                self.import_multisig(wallet)
            elif action == '6':
                self.transfer(wallet)
            elif action == '7':
                self.save_multisig_tx(wallet)
            elif action == '8':
                self.clear()
                print('Checking transaction...')
                self.show_multisig(wallet)
            elif action == '9':
                self.clear()
                print('Signing transaction...')
                self.sign_multisig(wallet)
            elif action == '10':
                self.clear()
                print('Submiting transaction...')
                self.submit_multisig(wallet)
            elif action == '11':
                self.clear()
                print('Fetching balance...')
                wallet.balance()
            elif action == '12':
                self.clear()
                print('Fetching transferts...')
                data = wallet.show_transfers()
            elif action == '13':
                self.clear()
                print('Rescaning blockchain...')
                wallet.rescan()
            elif action == '14':
                self.set_label(wallet)
            elif action == '15':
                wallet.run_console()
            elif action == '20':
                if input('Wallet directory will be archived and you will not see it in the list.\nYou can restore it later if you want.\n(Y/Yes/N/No): ').upper() in ('Y', 'YES'):  # noqa: E501
                    wallet.rpc_stop()
                    self.cd_default_dir()
                    if not os.path.isdir('archives'):
                        os.mkdir('archives')
                    fn = wallet.label or wallet.key1[:32]
                    tar = tarfile.open(os.path.join('archives', f'{fn}.tgz'), 'w:gz')
                    tar.add(os.path.join('wallets', wallet.name), arcname=fn)
                    tar.close()
                    shutil.rmtree(os.path.join('wallets', wallet.name))
                    print(f'Archive archives/{fn}.tgz added')
                    return True

    def get_wallets(self):
        self.cd_default_dir()
        paths = []
        for name in sorted(os.listdir('wallets')):
            if os.path.isdir(os.path.join('wallets', name)):
                paths.append(name)
        if not paths:
            print('No wallets found')
            return
        wallets = {}
        for p in paths:
            fn = os.path.join('wallets', p, 'ms.round1.txt')
            if os.path.exists(fn):
                key1 = open(fn).read().strip()
                if key1:
                    wallets[p] = {
                        'name': p,
                        'path': os.path.join('wallets', p),
                        'key1': f'{key1[:24]}..{key1[-24:]}',
                    }
                fn = os.path.join('wallets', p, 'wallet.label.txt')
                if os.path.exists(fn):
                    label = open(fn).read().strip()
                    if label:
                        wallets[p]['label'] = label
        return wallets

    def wallet_list(self):
        wallets = self.get_wallets()
        if not wallets:
            print('No wallets found')
            return
        while True:
            self.clear()
            print('Choose wallet:')
            for key, value in wallets.items():
                print(f'{key}. {value.get("label", value["key1"])}')
            print(f'{os.linesep}0. Back')
            action = input()
            self.clear()
            if action == '0':
                break
            wallet = wallets.get(action)
            if not wallet:
                print('Wallet not found, try again...')
                continue
            if self.wallet_detail(MultisigWallet(wallet['name'], self.settings)):
                wallets = self.get_wallets()
                if not wallets:
                    print('No wallets found')
                    return

    def search_wallet(self):
        self.cd_default_dir()
        print('Paste label, part of key1 or key2 or key3')
        while True:
            value = input()
            if not value:
                break
            if value.startswith('MultisigxV2R1'):
                fn = 'ms.round1.txt'
            elif value.startswith('MultisigxV2Rn1L'):
                fn = 'ms.round2.txt'
            elif value.startswith('MultisigxV2Rn1W'):
                fn = 'ms.round3.txt'
            elif len(value) == 95:
                fn = 'ms.address.txt'
            else:
                fn = 'wallet.label.txt'
            for name in sorted(os.listdir('wallets')):
                if os.path.isdir(os.path.join('wallets', name)):
                    filename = os.path.join('wallets', name, fn)
                    if os.path.exists(filename) and value in open(filename).read().strip():
                        self.wallet_detail(MultisigWallet(name, self.settings))
                        return
            print('Not found, try again...')

    def settings_page(self):
        self.clear()
        while True:
            print('Settings:')
            i = 1
            settings = {}
            for key, value in self.settings.items():
                print(f'{i}. {key} ({value})')
                settings[str(i)] = key
                i += 1
            print(f'{os.linesep}9. Test connection')
            print(f'{os.linesep}0. Back{os.linesep}')
            action = input()
            self.clear()
            if action == '0':
                break
            if action == '9':
                proxy = self.settings.get('proxy')
                if proxy:
                    proxy = f'socks5h://{self.settings["proxy"]}'
                try:
                    resp = requests.post(
                        urljoin(self.settings['daemon-address'], 'json_rpc'),
                        headers={'Content-Type': 'application/json'},
                        data=json.dumps({'jsonrpc': '2.0', 'id': 0, 'method': 'get_version'}),
                        verify=False,
                        proxies={
                            'http': proxy,
                            'https': proxy,
                        },
                        timeout=30
                    )
                    if resp.status_code == 200:
                        print(self.color(f'Server is available{os.linesep}', 'green'))
                    else:
                        print(self.color(resp.content.decode(), 'red'))
                except Exception as e:
                    print('Error:', e)
                continue
            key = settings.get(action)
            if key is None:
                print('Setting not found')
                continue
            value = self.settings.get(key)
            if value is None:
                print('Setting not found')
                continue
            if isinstance(value, bool):
                self.settings[key] = not self.settings[key]
            elif isinstance(value, str):
                print(f'Type new value for {key} or 0 to go back:')
                new_value = input()
                if new_value != '0':
                    self.settings[key] = new_value
            self.clear()
            self.save_settings()
            continue

    def bulk_generation(self, amount):
        print('Progress: 0%')
        keys = []
        time_started = time.time()
        for i in range(1, amount+1):
            wallet = MultisigWallet.generate_wallet(self.settings)
            wallet.turn_ms_on()
            wallet.prepare_multisig()
            keys.append(wallet.key1)
            wallet.rpc_stop()
            self.clear()
            percent = round(i * 100 / amount)
            time_passed = time.time() - time_started
            time_left = int(((amount - i) / i) * time_passed)
            print(f'Progress: {percent}%, Time left: {time_left}s')
            self.cd_default_dir()
        print(os.linesep.join(keys))

    def print_unused_keys(self):
        self.cd_default_dir()
        keys = []
        for name in os.listdir('wallets'):
            if os.path.isdir(os.path.join(self.default_path, 'wallets', name)):
                try:
                    wallet = MultisigWallet(name, self.settings)
                    if wallet.key1 and not wallet.key2:
                        keys.append(wallet.key1)
                except Exception:
                    print(self.color(f'Wallet #{name} is broken. If you do not need it, delete directory', 'red'))
            self.cd_default_dir()
        if not keys:
            print('Wallets not found')
        print(os.linesep.join(keys))

    def run(self):
        if not os.path.isdir('wallets'):
            os.mkdir('wallets')
        self.clear()
        while True:
            print('''
Choose action:
1. Generate new wallet
2. Wallet list
3. Search wallet

4. Settings

7. Print all unused round1 keys
8. Generate 10 new wallets
9. Generate 50 new wallets (can take very long time)

0. Exit
            ''')
            action = input()
            self.clear()
            if action == '0':
                exit()
            elif action == '1':
                print('Generating wallet...')
                wallet = MultisigWallet.generate_wallet(self.settings)
                wallet.turn_ms_on()
                wallet.prepare_multisig()
                self.cd_default_dir()
                self.wallet_detail(wallet)
            elif action == '2':
                self.wallet_list()
            elif action == '3':
                self.search_wallet()
            elif action == '4':
                self.settings_page()
            elif action == '7':
                self.print_unused_keys()
            elif action == '8':
                self.bulk_generation(10)
            elif action == '9':
                self.bulk_generation(50)
            else:
                print('Wrong action, try again...')


if __name__ == '__main__':
    Console().run()
